#!/usr/local/bin/perl

use subs qw( 
             Initialize 
             Parse ParseFunction
	     MCP
	     Domain GAMSIDS Delimit Gripe Controls Help
	     Clean Cleaner Cleanest
	     Alias Substitute Include Check Spliceificate
	     Decomment
	     Conclude
           );

#	Preprocess a file

my $GMF = Initialize;

$GMF = Parse $GMF;

Conclude $GMF;

#	Subroutines from here on

#	Global variables:  @ln, @ls, $alias, $GAMSID

sub Initialize {

$ARGV[0] && Help;

print STDERR <<"eop";

GAMS-F Function Preprocessor 12/04/98 11:07
by Collin Starkweather and Thomas F. Rutherford
-----------------------------------------------

Reading input file . . .
eop

$gmf = join '', <STDIN>;

unless ($gmf=~/\S/s) { Help }

study $gmf;

$alias = 0;

$GAMSID = '[a-zA-Z][a-zA-Z0-9_]{0,9}';

$gmf

} # end sub Initialize

sub Parse {

#	Parse a file

my ( $gmf, $f, $l ) = @_;

print STDERR "Parsing input file . . .\n\n" unless $f;

if ($gmf =~ /(FCN|DOM)\d{3}/is) {

	Gripe <<"eop";
The input file$ls[$l] contains an identifier of the form fcn???  or
dom???  where ?  represents a digit.  Identifiers of this form are
used by the function preprocessor for aliasing controlled sets, and
therefore are not permitted.

eop

}

my ( @gmf, $i, $head, $tail, $iteration, $statement, $stop );

@gmf = split /\n/, $gmf;

#	Scan the file looking for

#	1) Function definitions (and then comment them and insert any aliases)
#	2) Function references
#	3) MCP model formulations
#	4) Include files
#	5) The $SYSINCLUDE GAMS-F statement

$iteration = 0;

$ls[$l] = $f ? " (\U$f\E)" : '' ;

for ($ln[$l]=1,$i=0;$i<@gmf;$ln[$l]++,$i++) {

	#	Ignore comments and most dollar control operators

	while ( $gmf[$i]=~/^\*/
                or ( 
		     $gmf[$i]=~/^\$/ 
		     and 
		     $gmf[$i]!~/^(\$EXIT|\$STOP|\$ONTEXT|\$INCLUDE|\$SYSINCLUDE\s+GAMS-F)/i 
		   )
	      ) {

	        $ln[$l]++ , $i++

	}

	if ($gmf[$i] !~ /^\s*$/) {

	#	Make sure there is not more than one statement on the line

	splice @gmf, $i, 1, Spliceificate $gmf[$i], $l;

	if ($gmf[$i]=~/^(\$EXIT|\$STOP)/i) {

	#	Stop processing if an exit or stop is encountered

		last

	} elsif ($gmf[$i]=~/^\$ONTEXT/i) {

	#	Ignore $ONTEXT $OFFTEXT blocks

		while ($gmf[$i]!~/^\$OFFTEXT/i) { 
			$i++, $ln[$l]++;
			if ($i>$#gmf) {
				Gripe "\$ONTEXT encountered without a corresponding \$OFFTEXT!"
			}
		}

	} elsif ($gmf[$i]=~/==/) {

	#	The line is a function declaration

		if ($gmf[$i]=~/={3,}/) { Gripe "[$ln[$l]]\tToo many = signs$ls[$l]:", $gmf[$i] }

	#	Find the semicolon that ends the function declaration
	#	Eliminate any comments, blank lines, or $ontext $offtext blocks

		$stop = 0;
		until ($stop) {

			if ($gmf[$i]=~/;/s) { 
				$stop++ 
			} elsif ($gmf[$i+1]=~/^\*/) {
				splice @gmf, $i+1, 1; $ln[$l]++
			} elsif ($gmf[$i+1]=~/^\$ONTEXT/i) {
				until ($gmf[$i+1]=~/^\$OFFTEXT/i) {
					splice @gmf, $i+1, 1; $ln[$l]++
				}
				splice @gmf, $i+1, 1; $ln[$l]++
			} else {
				if ($i<$#gmf) {
					$gmf[$i] .= "\n" . $gmf[$i+1]; 
					splice @gmf, $i+1, 1;
					$ln[$l]++
				} else {
					Gripe "[$ln[$l]]\tMissing semicolon at end of file$ls[$l]?"
				}
			}

		}

	#	Make sure another statement doesn't follow the end of 
	#	the function declaration on the same line
		
		splice @gmf, $i, 1, Spliceificate $gmf[$i], $l;

		( $gmf[$i], $iteration ) = 
			ParseFunction $gmf[$i], $iteration, $l

	} elsif ($gmf[$i]=~/^\s*MODEL/ims) {

	#	The line is a MODEL declaration

	#	Find the semicolon that ends the function declaration
	#	Eliminate any comments, blank lines, or $ontext $offtext blocks

		$stop = 0;
		until ($stop) {

			Gripe "[$ln[$l]]\tMissing semicolon at end of file$ls[$l]?"
				if $i>$#gmf;

			if ($gmf[$i]=~/;/s) { 
				$stop++ 
			} elsif ($gmf[$i+1]=~/^\*/) {
				splice @gmf, $i+1, 1; $ln[$l]++
			} elsif ($gmf[$i+1]=~/^\$ONTEXT/i) {
				until ($gmf[$i+1]=~/^\$OFFTEXT/i) {
					splice @gmf, $i+1, 1; $ln[$l]++
				}
				splice @gmf, $i+1, 1; $ln[$l]++
			} else {
				if ($i<$#gmf) {
					$gmf[$i] .= "\n" . $gmf[$i+1]; 
					splice @gmf, $i+1, 1;
					$ln[$l]++
				} else {
					Gripe "[$ln[$l]]\tMissing semicolon at end of file$ls[$l]?"
				}
			}

		}

	#	Make sure another statement doesn't follow the end of 
	#	the function declaration on the same line
		
		splice @gmf, $i, 1, Spliceificate $gmf[$i], $l;

		$gmf[$i] = MCP $gmf[$i], $l;
			
	} elsif ($gmf[$i]=~/^\$INCLUDE/i) {

		$gmf[$i] = Include $gmf[$i], $l

	} elsif ($gmf[$i]=~/^(\$SYSINCLUDE\s+GAMS-F)/i) {

		$gmf[$i]=~s/^(\$SYSINCLUDE\s*GAMS-F)/*\t\1/i;
		$gmf[$i+1]=~s/^(\$EXIT)/*\t\1/i 
			if $i < @gmf and $gmf[$i+1]=~/^(\s*\$EXIT)/i

	} else {

	#	Look for function references

	#	Find the semicolon that ends the function declaration
	#	Eliminate any comments, blank lines, or $ontext $offtext blocks

		$stop = 0;
		until ($stop) {

			Gripe "[$ln[$l]]\tMissing semicolon at end of file$ls[$l]?"
				if $i>$#gmf;

			if ($gmf[$i]=~/;/s) { 
				$stop++ 
			} elsif ($gmf[$i+1]=~/^\*/) {
				splice @gmf, $i+1, 1; $ln[$l]++
			} elsif ($gmf[$i+1]=~/^\$ONTEXT/i) {
				until ($gmf[$i+1]=~/^\$OFFTEXT/i) {
					splice @gmf, $i+1, 1; $ln[$l]++
				}
				splice @gmf, $i+1, 1; $ln[$l]++
			} else {
				if ($i<$#gmf) {
					$gmf[$i] .= "\n" . $gmf[$i+1]; 
					splice @gmf, $i+1, 1;
					$ln[$l]++
				} else {
					Gripe "[$ln[$l]]\tMissing semicolon at end of file$ls[$l]?"
				}
			}

		}

	#	Make sure another statement doesn't follow the end of 
	#	the function declaration on the same line
		
		splice @gmf, $i, 1, Spliceificate $gmf[$i], $l;

		for $function (@function) {

			if ($gmf[$i]=~/$function->{key}/s) {

				$gmf[$i] = Substitute $function, $gmf[$i], $GAMS=1, $l;

			}
		}
	}

	}

}

print STDERR "\tNo functions detected.\n" unless @function;

$gmf = join "\n", @gmf;

$gmf

} # end sub Parse

sub Spliceificate {

#	If a line contains more than one statement, split it into
#	several lines

my ($tail, $l) = @_;

my ( @list, $head );

while ($tail=~/;(\s*\S)+\s*$/) {

#	We have more than one statement on the line 
#	Split the line at the ;

#	Strange:  g only works here in a control expression

	if ($tail=~/;/g) {

		$head = substr $tail, 0, pos $tail;
		$tail = substr $tail, pos ($tail);

		push @list, $head;

		pos ($tail) = 0;

		$ln[$l]--

	}

}

push @list, $tail;

@list

}

sub ParseFunction {

my ( $statement, $iteration, $l ) = @_;

my ( $LHS, $RHS, $key, $name, $domain, $value, $alias, $aliases, $head, $tail );

$statement =~ /(.*)==(.*);/s;

( $LHS ) = Cleanest $1;
$RHS = $2;

if ($LHS=~/=/s or $RHS=~/=/s) { 

	Gripe "[$ln[$l]]$ls[$l] Did you forget a semicolon close to the definition of ", "$LHS?" 

}

#	Get the name and domain (i.e., the sets over which the function is defined)

$LHS =~ /^($GAMSID)(\(.*?\))?$/s;

( $name, $domain ) = Cleaner ( $1, $2 );

#	Make a suitable regex pattern for finding function references
if ($domain) {
	( $key = $LHS ) =~ s/^($GAMSID)(\().*$/(?i)\\b\1\\s*\\\2/;
} else {
	$key = '(?i)\b'.$name.'\b'
}

#	Strip any parentheses off the domain declaration

$domain =~ s/^\((.*)\)$/\1/s;

#	Create aliases for any controlled sets on the RHS; e.g., 
#	SUM((I,J), . . . will have aliases SUM((FCN000,FCN001), . . .
( $value, $alias, $aliases ) = Alias $RHS, $iteration;

#	Substitute any previously declared functions into the RHS
for ($j=0;$j<$iteration;$j++) { 
	if (@function) {
		$value = Substitute $function[$j], $value, 0, $l
	}
}

#if ($name =~ /v/i) { die "\nValue is \n\n$value\n\n" }

#	Take any superfluous whitespace out of $value
$value =~ s/\s+/ /gs;

#	GAMS V2.25 will only accept 4096 characters per line, so chop up
#	$value if necessary
if (length ($value) > 4000) {
	$head = '';
	$tail = $value;
	while (length ($tail) > 4000) {
		pos($tail) = 4000;
		$tail =~ /\b($GAMSID)\b/g;
		$head = substr($tail,0,pos($tail)-length($1)) . "\n";
		$tail = "\t" . substr($tail,pos($tail)-length($1));
		$tail = "\t$tail"
	}
	$value = $head . $tail
}

my @domain = Domain $domain;

#	The original value is used for any MCP replacements in the 
#	declaration EQN???(o).. F(o) =E= 0;
my $original_value = $value;

#	Replace domain references with tokens for easier substitutions 
#	later on
$i=0;
my @pattern = ();
for (@domain) {
	( $pattern = $_ ) =~ s/(\W)/\\\1/g;
	$pattern =~ s/^(\w+)/\\b\1/;
	$pattern =~ s/(\w+)$/\1\\b/;
	$pattern[$i++] = $pattern
}
$i = 0;
for (@pattern) {
	$token[$i] = sprintf "#DOM%03d#", $i;
	$value =~ s/([^\$])$_/\1$token[$i]/g;
	$i++
}

#	Add the function data type to the function stack
push @function, { 
	'name' => $name, 
	'key' => $key,
	'value' => $value,
	'alias' => $aliases,
	'original_value' => $original_value,
	'token' => [@token],
	'domain_name' => $domain, 
	'domain' => [@domain], 
		};

#	Comment out the function statement and insert any aliases
$statement =~ s/^(.)/*\t\1/gm;
if ($alias) {

$statement .= <<"eop";

*--------------------------------------------------------
*	The following aliases were created as a result of 
*	function \U$name\E (defined above)
*
$alias
eop

}
print STDERR "\t[$ln[$l]]\tFunction object \U$name\E compiled$ls[$l]\n";

$iteration++;

( $statement, $iteration )

} # end sub ParseFunction

sub Substitute {

#	Substitute a function value into a string

my ( $function, $string, $GAMS, $l ) = @_;

my ( $head, $tail, $rvalue, $value, $body, $domain, @ref_domain );

#	Find a reference to the function
my $substitute_found = 0;

if ($string=~/$function->{key}/s) { 

	$substitute_found = 1;

	$rvalue = $function->{value};

	#	Only substitute in aliases if needed

	for ( keys %{$function->{alias}} ) {

		$rvalue=~s/$_/$function->{alias}->{$_}/gis if $string=~/$_/s

	}

}

( $head, $tail ) = ( '', $string );

while ($tail=~/$function->{key}/gs) {

	print STDERR "\t[$ln[$l]]\tPerforming substitution of function \U$function->{name}\E$ls[$l]\n";

	$value = $rvalue;
	( $head .= substr $tail, 0, pos $tail ) =~ s/$function->{key}$//;
	$tail = substr $tail, pos $tail;
	if ($function->{key}=~/\($/) {
		( $body, $tail ) = Delimit $tail, ')';
		( $body ) = Clean $body;
		( $domain = $body ) =~ s/\)$//;
		@ref_domain = Domain $domain;
		if (@ref_domain != @{$function->{'domain'}}) {
			Gripe "[$ln[$l]] Wrong number of domain arguments ($domain) in function ".
                              "$function->{name} reference$ls[$l]!";
		}
		for ($i=0;$i<@{$function->{token}};$i++) { 
		#	Eliminate any multiplication or exponentiation operators
		#	that might appear as the first character(s) of a line
			$value =~ s/($function->{token}->[$i])(\*{1,2})/\1\t\2/g 
				if $ref_domain[$i]=~/\n$/s;
		#	Substitute in the values of the function domain
			$value =~ s/$function->{token}->[$i]/$ref_domain[$i]/g ;
		}
	}

	if ($GAMS) {

		$body =<<"eop";


*	#### Preprocessor function replacement for \U$function->{name}\E ####

(
$value
)

eop

	#	Make sure that a multiplication or exponentiation operator 
	#	might not come out placed as a comment
		$tail =~ s/^(\*{1,2})/\t\1/s;
	#	Check for dollar controls (e.g., CESCOST$FUNCTION("CES"))
		if ($tail =~ /^\$/) { $body =~ s/\n+$//s }

	} else {

		$body = "($value)";

	}

$head .= $body;

pos ($tail) = 0;

} # end while

$head . $tail;

} # end sub Substitute

#	Remove any superfluous whitespace that is not a Newline from a list
sub Clean { my @l=@_; for (@l) { s/[ \r\t\f]+/ /gs } @l }

#	Remove any superfluous whitespace from a list
sub Cleaner { my @l=@_; for (@l) { s/\s+/ /gs } @l }

#	Remove any whitespace from a list
sub Cleanest { my @l=@_; for (@l) { s/\s+//gs } @l }

sub Alias {

my ( $expression, $iteration ) = @_;

my ( $head, $tail, @controls, $control, $FCN, $string, %aliases );

#	Identify all additional controlled sets

$head = '';
$tail = $expression;

@controls = ();

while ($tail =~ /\bsum\b|\bprod\b|\bsmin\b|\bsmax\b/is) {

	if ($tail=~/\bsum\s*\(|\bprod\s*\(|\bsmin\s*\(|\bsmax\s*\(/gis) {

	#	Separate the body of the sum etc. statement from the rest
	#	and get the domains (which may include sets or dynamic sets)
	#	controlled by the sum etc.

		$head .= substr $tail, 0, pos $tail;
		$tail = substr $tail, pos $tail;

		( $control, $tail ) = Delimit $tail, ',';

		@controls = ( @controls, Controls $control );

		$head .= $control;

		pos ($tail) = 0;

		}

}

$expression = $head . $tail;

for $control (@controls) { 

	$FCN = sprintf "FCN%03d", $alias++;
	$aliases{'\b'.$control.'\b'} = $FCN;
	#	Generate a string to pass back to the function 
	$string .= "ALIAS($FCN,$control);\n";

}

( $expression, $string, \%aliases )

} # end sub Alias

sub Conclude {

my $gmf = shift;

$gmf .= "\n";

print STDERR "\nWriting output . . .\n";

print STDOUT $gmf;

print STDERR "\nGAMS-F terminating normally.\n\n";

exit 1

} # end sub Conclude

sub Domain {

#	Given a string of domains (e.g., sets, dynamic sets, parameters, variables) 
#	separated by commas, returns a list of the domains; e.g., 
#	the domain of F(I(J),K) is the list I(J) K

my $domain = shift;

( $domain ) = Clean $domain;

my $pos = 0;

my ( $item, $count, $char, @domain );

while ($pos < length $domain) {

	$item = '';
	$count = 1;
	while ($count) {

		if ( $pos > length $domain ) {
			Gripe "Comma expected at expression $domain ", $item
		}

		$char = substr $domain, $pos++, 1;

		$count-- if $char=~/,|\)/ or $pos == length $domain;

		if ($char eq '(') {

			$substr = substr $domain, $pos;
			( $head, $tail ) = Delimit $substr, ')';
			$item .= $char . $head;
			$pos = $pos + length $head;
			$count-- if $pos == length $domain;

		} else {

			$item .= $char

		}

	}

	#	Get rid of the trailing comma
	$item =~ s/,$//;

	#	Get rid of any dollar control operators
	$item =~ s/\$.*$//;

	#	Push the item on the domain stack
	push @domain, $item

}

@domain

} # end sub Domain

sub Controls {

my $domain = shift;

my (@controls,@dynamicsets);

#	Any GAMS ID that does not appear in front of a dollar control operator
#	or a GAMS ID representing a dynamic set is considered a control; e.g.,
#	
#	Domain			Controls
#	(i)			i
#	((i,j),k)		i j k
#	(i,map(j,k),l)		i map l
#	(i$j(i))		i
#	(i$j(i),map(j,k))	i map

@controls = Domain $domain;

my @return = ();
my @push = ();

for (@controls) { 

#	Strip off any dollar control operators
	s/(.*)\$.*/\1/;

#	Identify any dynamic sets
	s/\b($GAMSID)\b\s*\(.*/\1/;

	@push = /\b$GAMSID\b/g;

	push @return, @push

}

@return

} # end sub Controls

sub Delimit {

my ( $string, $delimiter, $CLEAN ) = @_;

#	Finds the next $delimiter, ignoring anything in between that is not
#	$delimiter or that is enclosed in parentheses.  Returns
#	$head and $tail of $string corresponding to $string before 
#	and including $delimiter and $string after the $delimiter

( $string ) = Clean ($string) if $CLEAN;

my $char = '';
my $pos = 0;
my ( $count, $head, $tail );

until ($char eq $delimiter) {

	$char = substr $string, $pos++, 1;

	$head .= $char;

	if ($char eq '(') {

		$count = 1;

		while ($count) {

			$char = substr $string, $pos++, 1;
			$count++ if $char eq '(';
			$count-- if $char eq ')';
			if ($pos + 1 > length $string) {
				Gripe "Unmatched delimiter detected at ", $head . $char
			} else {
				$head .= $char
			}
			$char = ''
		}
	}
}

$tail = substr $string, $pos;

( $head, $tail )

} # end sub Delimit

sub Gripe {

my ( $message, $string, $text, $help ) = @_;

if ($string) { $string = "\n\n$string" }
die <<"eop";

Error:

$message $string

Processing aborted.

eop

} # end sub Gripe

sub GAMSIDS {

my $domain = shift;

my @GAMSIDS;

#	Identifies all valid GAMS identifiers in a string

@GAMSIDS = $domain =~ /\b$GAMSID\b/g;

@GAMSIDS

} # end sub GAMSIDS

sub MCP {

#	Substitute a function into a model declaration for MCP compatibility

my $statement = shift;

my $in;

print STDERR "\t[$ln[$l]]\tChecking MODEL statement$ls[$l] . . . \n";

my $substitution = 0;

( $in = $statement ) =~ s/^\s*//;

if ($in=~/^model\s*$GAMSID/i) { 
	$in=~s/^(model\s*$GAMSID).*/\1/is
} else {
	Gripe "[$ln[$l]] Check model statement syntax$ls[$l]:  ", $in
}

for $function (@function) {

	if ($statement=~/^[^;]*(\b$function->{name}\b)(\s*\.\s*$GAMSID\s*)(,)?/gis) {

		my $variable = $2;
		my $comma = $3;

		my $clean_variable;
		( $clean_variable ) = Clean ( $variable );
		my $nodot = $clean_variable;
		$nodot =~ s/\W//g;

		$substitution = 1;

		$eqno = sprintf "%03d", $eqn;
		$eqn++;

		my $domain = $function->{domain_name} ? "($function->{domain_name})" : "";

$string .= <<"eop";

EQUATION EQN$eqno Defines \U$function->{name}\E complementary to \U$nodot\E in \U$in\E;

EQN$eqno$domain..	$function->{original_value} =E= 0;

eop

my $insert = <<"eop";


*	MCP function substitution:  \U$function->{name}$clean_variable\E replaced by 

EQN$eqno$variable

eop

$insert .= "\t$comma";

		print STDERR "\t[$ln[$l]]\tMCP conditions formulated for \U$function->{name}\E$ls[$l] . . . \n";

		substr ($statement, 
			pos ($statement)-length($function->{name})-length($variable)-length($comma),
			length($function->{name})+length($variable)+length($comma)) = $insert;
	}

}

print STDERR "\t\tNo function complementarity found\n" unless $substitution;

$string . $statement;

} # end sub MCP

sub Help {

print <<"eop";


GAMS-F Function Preprocessor 12/04/98 11:07
by Collin Starkweather and Thomas F. Rutherford
===============================================

Brief Overview
--------------

This tool allows for the definition of functions within the body of a
GAMS file which will later be substituted into parameter or equation
definitions.  Thus, a function may be defined in a GAMS file as

F(I) == A * K(I)**alpha * L(I)**(1-alpha)

and later referenced in an equation as

OUTPUT(I)..	Y(I) =E= F(I);

Typically, GAMS-F would be used with the \$sysinclude syntax given
below.

The Details
-----------

Command line syntax:  	gams-f <input_file >output_file

Translates a GAMS input file with function definitions to a standard
GAMS program in which function definitions are omitted and function
references are replaced.  The input file comes from standard input
and output is written to standard output.  Passing a command-line 
argument to GAMS-F will result in this message.

Syntax in GAMS:		\$sysinclude gams-f

Press Enter or Return to continue, Ctrl-C to exit
eop

<STDIN>;

print <<"eop";
Function Definitions
--------------------

<func id>(args) == expression(args);

where the argument list may include sets, parameters or variables.

Here are a few examples:

(i)     F(i) == SUM(j, alpha(i,j) * 
                        P(j)**(1-sigma))**(1/(1-sigma));

(ii)    G(i,j) == PROD(t(k), a(i,j,k) * X(i,j)**alpha);

(iii)   H(i) == SMIN(j\$(alpha(i) gt alpha(j)), a(i,j) * F(j));

The preprocessor reads these function definitions and generates GAMS
code which aliases any dummy sets and then makes the appropriate
substitutions when the functions are referenced later in the code.
Note that function definitions can reference previously defined
functions. 

NB: The preprocessor does not perform domain checking.  Apart from
counting dimensions in function references, all syntax checking is
performed by the GAMS compiler on the translated code.

Press Enter or Return to continue, Ctrl-C to exit
eop

<STDIN>;

print <<"eop";
Function References
-------------------
A function may only be referenced following its definition.

Functions may be referenced in parameter assignments, equations or
in PUT statements.  Functions may not appear in DISPLAY statements.

Functions may be used in place of an equation name in the MODEL
statement for a mixed complementarity problem.  For example, 

F(i) == q(i) + sum(j, M(i,j) * Z(j));

model lemke / F.Z /;

defines a linear complementarity problem, LCP(q,M).  

Restrictions
------------
o The function identifier and arguments must appear on the same line
  as the assignment symbol (==), however the function definition may
  span multiple lines.

o Function references appearing within \$ontext \$offtext blocks are 
  ignored.

o \$include files will be included by the preprocessor only if 

  1)  They include function definitions,
  2)  They include references to previously defined functions, or
  3)  They \$include a file which (may include a file which . . .)
      meets criteria (1) or (2)

  The \$include statement may not include %symbols%; i.e., 

  \$include "data/%data%.txt"

  would return an error.  

o GAMSF uses set names FCN000, FCN001 etc. for sets and names EQN000,
  EQN001, etc. for MCP equations.  Thus, GAMS identifiers of the form 
  FCN??? or EQN??? are not permitted.

Press Enter or Return to continue, Ctrl-C to exit
eop

<STDIN>;

print <<"eop";
More Documentation
------------------
In the spirit of learning by doing, we have put together a set of
sample programs illustrating the use of functions in GAMS programs.
See http://robles.colorado.edu/~tomruth/inclib/gams-f.htm.

Credits and Complaints
----------------------
This application was designed and implemented by Thomas Rutherford
and Collin Starkweather in the Department of Economics at the
University of Colorado, Boulder.  If you encounter problems or have
suggestions for additional features, please contact
collin.starkweather\@colorado.edu.  

This research supported by the GAMS Applied General Equilibrium Research
Fund. The authors remain responsible for any bugs which exist in this
software. This software is not officially supported by GAMS Corporation. 

eop

exit 1;

} # end sub Help

sub Include {

#	Include a file if 

#	1)	The file contains a function reference, 
#	2)	The file contains a reference to a previously defined function, or
#	3)	One of the nested includes does

#	Takes a filename as an argument

my ( $statement, $l ) = @_;

my ( @list, $i, $f, $include, $check );

#	Remove any quotes from the filename

$statement =~ /\$INCLUDE\s+(.*)$/i;

( $f )  = Clean $1;

$f =~ s/^['"](.*)['"]$/\1/;

$f .= '.gms' if -f $f . '.gms';

if ( Check $f, $l ) {

	print STDERR "\t[$ln[$l]]\tIncluding file $f$ls[$l] . . .\n";
	open (F, $f) || warn "\n\t[$ln[$l]]\tCan't open include file $f:  $!$ls[$l]\n\n";
	$include = join '', <F>;
	close F;
	$statement = &Parse ($include,$f,$l+1)

}

$statement

} # end sub Include

sub Check {

my ($f, $l) = @_;

my ( @list, $i, $include, $check, $found, @include );

#	Remove any quotes from the filename

( $f ) = Clean $f;

$f =~ s/^['"](.*)['"]$/\1/;

$f .= '.gms' if -f $f . '.gms';

print STDERR "\t[$ln[$l]]\tExamining file $f$ls[$l] . . .\n";

$found = 0;

if (open (F, $f)) {

	$include = join '', <F>;
	close F;

	#	Scan the include file for any function definitions or references

	#	Eliminate comments and blank lines
	$check = $include;
	study $check;
	$check =~ s/^\*.*$//mg;
	$check =~ s/\$ONTEXT.*\$OFFTEXT//igs;
	$check =~ s/^\$(?!\bINCLUDE\b).*$//gim;
	$check =~ s/\n\s*\n/\n/g;

	#	Only include the file if it has a function definition
	#	or references a previously defined function or includes 
	#	another file

	#	Check for function definitions

	$found = 1 if $check =~ /^\$INCLUDE|[^=]==[^=]/ims;

	if (not $found) {

	#	Check for function references

		for $function (@function) { $found=1 if $check=~/$function->{key}/s }

	}

	if (not $found) {

	#	Check for $INCLUDE statements

		my @include = split /\n/, $include;

		for (@include) {

			if (/^\$INCLUDE\s*(.*)/) {
print STDERR "Found \$INCLUDE\n";
				$found = &Check($1,$l)

			}

		}


	}

} else {

	 warn "\n\t[$ln[$l]]\tCan't open include file $f:  $!$ls[$l]\n\n"

}

$found

}

sub Decomment {

#	Eliminates any comments or $ontext $offtext blocks from a string

my $string = shift;

$string =~ s/^\*.*$//gm;
$string =~ s/\$ontext.*?\$offtext//gs;

$string

}
